
/*!
 *************************************************************************************
 * \file adaptive_loop_filter.c
 *
 * \brief
 *    adaptive loop filter routines
 *
 * \author
 *    Takeshi Chujoh         <takeshi.chujoh@toshiba.co.jp>             \n
 *    Akiyuki Tanizawa       <akiyuki.tanizawa@toshiba.co.jp>           \n
 *    Goki Yasuda            <goki1.yasuda@toshiba.co.jp>               \n
 *    Naofumi Wada           <naofumi.wada@toshiba.co.jp>               \n
 *    Takashi Watanabe       <takashi39.watanabe@toshiba.co.jp>         \n
 *************************************************************************************
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
#include <float.h>

#include "adaptive_loop_filter.h"
#include "image.h"
#include "memalloc.h"
#include "vlc.h"
#include "cabac.h"
#include "header.h"

#ifdef ADAPTIVE_LOOP_FILTER

//------- global parameters for adaptive_loop_filter.c
int    num_bit;
int    r_offset;
int    ext_offset;
int    base_type;
double lambda_alf_luma;
double lambda_alf_chroma;
int64  rate, min_rate, base_rate, orig_rate;
int64  dist, min_dist, base_dist, orig_dist;
double cost, min_cost, base_cost, orig_cost;
int64  tmp_luma_rate;

alf_parameter_set_rbsp_t *temp_alfps;
double ***correlation;
double *dbl_alf_coef;
imgpel **imgY_rest;
imgpel ***imgUV_rest;
imgpel **imgY_ext;
imgpel ***imgUV_ext;
imgpel **imgY_temp;
imgpel **imgY_best;
int    **alf_coef_base;
int    ***alf_coef_tap_base;
int    **alf_coef_base_chroma;
Bitstream *tempStream;
Bitstream *alfstream;
unsigned int ****blk_corr;
unsigned int ****blk_corr5x5;
unsigned int ****blk_corr7x7;

int    alf_blk_size[8] = { 8, 16, 24, 32, 48, 64, 96, 128 };

int filter_base_luma[42] = {
  0,  0,  0,  0,   0,  0,  0,  0, 0,
  0,  0,  0, -1,   5, -1,  0,  0, 0,
  0,  0,  0,  0,  -7,  0,  0,  0, 0,
  0, -1,  0, -5,  18, -5,  0, -1, 0,
  0,  5, -7, 18, 219,  0
};
int filter_base_chroma[14] = {
   5, -5,  -5, -5,  5,
  -5,  0,  40,  0, -5,
  -5, 40, 160,  0
};

int filter_pos_5x5_in_9x9[14] = {
  20, 21, 22, 23, 24,
  29, 30, 31, 32, 33,
  38, 39, 40, 41
};
int filter_pos_7x7_in_9x9[26] = {
  10, 11, 12, 13, 14, 15, 16,
  19, 20, 21, 22, 23, 24, 25,
  28, 29, 30, 31, 32, 33, 34,
  37, 38, 39, 40, 41
};
int filter_pos_9x9_in_9x9[42] = {
   0,  1,  2,  3,  4,  5,  6,  7,  8, 
   9, 10, 11, 12, 13, 14, 15, 16, 17,
  18, 19, 20, 21, 22, 23, 24, 25, 26,
  27, 28, 29, 30, 31, 32, 33, 34, 35,
  36, 37, 38, 39, 40, 41
};
int filter_pos_5x5_in_7x7[14] = { 
   8,  9, 10, 11, 12,
  15, 16, 17, 18, 19,
  22, 23, 24, 25
};

int symmetric_array9x9[81] = {
   0,  1,  2,  3,  4,  5,  6,  7,  8,
   9, 10, 11, 12, 13, 14, 15, 16, 17,    
  18, 19, 20, 21, 22, 23, 24, 25, 26,
  27, 28, 29, 30, 31, 32, 33, 34, 35,       
  36, 37, 38, 39, 40, 39, 38, 37, 36, 
  35, 34, 33, 32, 31, 30, 29, 28, 27, 
  26, 25, 24, 23, 22, 21, 20, 19, 18,
  17, 16, 15, 14, 13, 12, 11, 10,  9, 
   8,  7,  6,  5,  4,  3,  2,  1,  0
};
int symmetric_array7x7[49] = {
   0,  1,  2,  3,  4,  5,  6,    
   7,  8,  9, 10, 11, 12, 13,
  14, 15, 16, 17, 18, 19, 20,     
  21, 22, 23, 24, 23, 22, 21, 
  20, 19, 18, 17, 16, 15, 14, 
  13, 12, 11, 10,  9,  8,  7,
   6,  5,  4,  3,  2,  1,  0, 
};
int symmetric_array5x5[25] = {    
   0,  1,  2,  3,  4, 
   5,  6,  7,  8,  9,        
  10, 11, 12, 11, 10,  
   9,  8,  7,  6,  5,  
   4,  3,  2,  1,  0,  
};

// for quantization of filter coef.
int symmetric_mag9x9[41] = {
  2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 1
};
int symmetric_mag7x7[25] = {
  2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 1
};
int symmetric_mag5x5[13] = {
  2, 2, 2, 2, 2,
  2, 2, 2, 2, 2,
  2, 2, 1
};

int diff_pos_5x5_in_7x7[12] = { 0, 1, 2, 3, 4, 5, 6, 7, 13, 14, 20, 21 };
int same_pos_5x5_in_7x7[13] = {
   8,  9, 10, 11, 12,
  15, 16, 17, 18, 19,
  22, 23, 24
};
int diff_pos_7x7_in_9x9[16] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 17, 18, 26, 27, 35, 36};
int same_pos_7x7_in_9x9[25] = {
  10, 11, 12, 13, 14, 15, 16,
  19, 20, 21, 22, 23, 24, 25,
  28, 29, 30, 31, 32, 33, 34,
  37, 38, 39, 40
};
int filter_base_flag[3][4];
//------- end



/*! 
*************************************************************************************
* \brief
*    Allocates memory for an alfps
*
* \return
*    pointer to a alfps
*************************************************************************************
*/
alf_parameter_set_rbsp_t *AllocALFPS (int *memory_size)
{
  alf_parameter_set_rbsp_t *p;

  if ((p=calloc (sizeof (alf_parameter_set_rbsp_t), 1)) == NULL)
    no_mem_exit ("AllocALFPS: ALFPS");
  *memory_size += sizeof(alf_parameter_set_rbsp_t);

  *memory_size += get_mem1Dint(&(p->alf.coeff), MAX_NUM_COEF);
  *memory_size += get_mem1Dint(&(p->alf.coeff_chroma), MAX_NUM_COEF_C);

  if(((p->alf_blk_data) = (AlfBlock *) calloc(img->max_num_alf_block, sizeof(AlfBlock))) == NULL)
    no_mem_exit("AllocALFPS: p->alf_blk_data");
  *memory_size += sizeof(AlfBlock)*img->max_num_alf_block;

  return p;
}

/*! 
*************************************************************************************
* \brief
*    Frees an alfps
*
* \param alfps
*     alfps to be freed
*
* \return
*    none
*************************************************************************************
*/
void FreeALFPS (alf_parameter_set_rbsp_t *p)
{
  assert(p != NULL);

  if(p->alf.coeff != NULL)
  {
    free_mem1Dint (p->alf.coeff);
    p->alf.coeff=NULL;
  }

  if(p->alf.coeff != NULL)
  {
    free_mem1Dint (p->alf.coeff_chroma);
    p->alf.coeff_chroma=NULL;
  }

  if(p->alf_blk_data != NULL)
  {
    free(p->alf_blk_data);
    p->alf_blk_data=NULL;
  }

  free (p);
  p=NULL;
}

/*!
************************************************************************
* \brief
*    Dynamic memory allocation of frame size related global buffers
*    buffers are defined in global.h, allocated memory must be freed in
*    void FreeALFGlobalBurrers()
*
*  \par Output:
*     Number of allocated bytes
***********************************************************************
*/
int  InitALFGlobalBuffers(void)
{
  int memory_size = 0;
  
  memory_size += get_mem3Ddouble (&(correlation),  3, MAX_NUM_COEF, MAX_NUM_COEF+1);
  memory_size += get_mem1Ddouble (&(dbl_alf_coef), MAX_NUM_COEF);

  memory_size += get_mem2Dpel (&(imgY_rest), img->height, img->width);
  memory_size += get_mem3Dpel (&(imgUV_rest), 2, img->height_cr, img->width_cr);
  memory_size += get_mem2Dpel (&(imgY_ext),  img->height+MAX_NUM_TAP, img->width+MAX_NUM_TAP);
  memory_size += get_mem3Dpel (&(imgUV_ext),  2, img->height_cr+MAX_NUM_TAP_C, img->width_cr+MAX_NUM_TAP_C);
  memory_size += get_mem2Dint (&(alf_coef_base), 4, MAX_NUM_COEF);
  memory_size += get_mem3Dint (&(alf_coef_tap_base), 3, 4, MAX_NUM_COEF);
  memory_size += get_mem2Dint (&(alf_coef_base_chroma), 4, MAX_NUM_COEF_C);
  memory_size += get_mem2Dpel (&(imgY_temp), img->height, img->width);
  memory_size += get_mem2Dpel (&(imgY_best), img->height, img->width);

  // for block-based filtering
  img->max_WidthInAlfBlks  = img->width  / alf_blk_size[0];
  img->max_HeightInAlfBlks = img->height / alf_blk_size[0];
  if( (img->width  % alf_blk_size[0])!=0 ) img->max_WidthInAlfBlks++;
  if( (img->height % alf_blk_size[0])!=0 ) img->max_HeightInAlfBlks++;
  img->max_num_alf_block   = img->max_WidthInAlfBlks * img->max_HeightInAlfBlks;
  memory_size += get_mem_blk_corr (&(blk_corr),    img->max_HeightInAlfBlks, img->max_WidthInAlfBlks, MAX_NUM_COEF, MAX_NUM_COEF+1);
  memory_size += get_mem_blk_corr (&(blk_corr5x5), img->max_HeightInAlfBlks, img->max_WidthInAlfBlks, NUM_COEF_5x5, NUM_COEF_5x5+1);
  memory_size += get_mem_blk_corr (&(blk_corr7x7), img->max_HeightInAlfBlks, img->max_WidthInAlfBlks, NUM_COEF_7x7, NUM_COEF_7x7+1);

  // allocate bitstream
  if ( (alfstream =  (Bitstream *) calloc(1, sizeof(Bitstream))) == NULL )
    no_mem_exit ("InitALFGlobalBuffers: alfstream");
  memory_size += sizeof(Bitstream);
  if ( (alfstream->streamBuffer = (byte *) calloc(MAX_ALF_PARAM_SIZE, sizeof(byte))) == NULL )
    no_mem_exit ("InitALFGlobalBuffers: alfstream->streamBuffer");
  memory_size += MAX_ALF_PARAM_SIZE*sizeof(byte);
  alfstream->bits_to_go=8;
  alfstream->byte_buf=0;
  alfstream->byte_pos=0;

  // allocate ALF parameter set
  temp_alfps   = AllocALFPS(&memory_size);
  active_alfps = AllocALFPS(&memory_size);

  return memory_size;
}

/*!
************************************************************************
* \brief
*    Free allocated memory of frame size related global buffers
*    buffers are defined in global.h, allocated memory is allocated in
*    int InitALFGlobalBuffers()
************************************************************************
*/
void FreeALFGlobalBurrers(void)
{
  free_mem3Ddouble (correlation, 3);
  free_mem1Ddouble (dbl_alf_coef);

  free_mem2Dpel (imgY_rest);
  free_mem3Dpel (imgUV_rest, 2);
  free_mem2Dpel (imgY_ext);
  free_mem3Dpel (imgUV_ext, 2);
  free_mem2Dint (alf_coef_base);
  free_mem3Dint (alf_coef_tap_base, 3);
  free_mem2Dint (alf_coef_base_chroma);

  free_mem_blk_corr (blk_corr,    img->max_HeightInAlfBlks, img->max_WidthInAlfBlks);
  free_mem_blk_corr (blk_corr5x5, img->max_HeightInAlfBlks, img->max_WidthInAlfBlks);
  free_mem_blk_corr (blk_corr7x7, img->max_HeightInAlfBlks, img->max_WidthInAlfBlks);

  if(alfstream->streamBuffer != NULL)
  {
    free(alfstream->streamBuffer);
    alfstream->streamBuffer=NULL;
  }
  if(alfstream != NULL)
  {
    free(alfstream);
    alfstream=NULL;
  }

  free_mem2Dpel    (imgY_temp);
  free_mem2Dpel    (imgY_best);

  FreeALFPS (temp_alfps);
  FreeALFPS (active_alfps);
}

/*!
 ************************************************************************
 * \brief
 *    set ALF parameters (p -> alfps)
 ************************************************************************
 */
void SetALFParameters( alf_parameter_set_rbsp_t *alfps, alf_parameter_set_rbsp_t *p )
{
  alfps->alf_flag              = p->alf_flag;

  if(alfps->alf_flag)
  {
    alfps->alf.tap               = p->alf.tap;
    alfps->alf.num_coeff         = p->alf.num_coeff;
    memcpy(alfps->alf.coeff, p->alf.coeff, sizeof(int)*(alfps->alf.num_coeff));
    alfps->chroma_idc            = p->chroma_idc;
    if(alfps->chroma_idc)
    {
      alfps->alf.tap_chroma        = p->alf.tap_chroma;
      alfps->alf.num_coeff_chroma  = p->alf.num_coeff_chroma;
      memcpy(alfps->alf.coeff_chroma, p->alf.coeff_chroma, sizeof(int)*(alfps->alf.num_coeff_chroma));
    }
    alfps->block_control_flag    = p->block_control_flag;
    if(alfps->block_control_flag)
    {
      alfps->alf_block_size_idx  = p->alf_block_size_idx;
      alfps->alf_block_size      = p->alf_block_size;
      alfps->WidthInAlfBlks      = p->WidthInAlfBlks;
      alfps->HeightInAlfBlks     = p->HeightInAlfBlks;
      alfps->num_alf_block       = p->num_alf_block;
      memcpy( alfps->alf_blk_data, p->alf_blk_data, sizeof(AlfBlock)*(alfps->num_alf_block));
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    make temporally frame for filtering
 *****************************************************************************************
 */
void MakeExtendFrame(imgpel** ImgDec, imgpel** ImgExt, int width, int height, int offset)
{
  int x, y;

  for(y=0; y<offset; y++)
  {
    for(x=0; x<offset; x++)
    {
      ImgExt[y][x] = ImgDec[0][0];
    }
    for(x=offset; x<width+offset; x++)
    {
      ImgExt[y][x] = ImgDec[0][x-offset];
    }
    for(x=width+offset; x<width+(offset<<1); x++)
    {
      ImgExt[y][x] = ImgDec[0][width-1];
    }
  }
  for(y=offset; y<height+offset; y++)
  {
    for(x=0; x<offset; x++)
    {
      ImgExt[y][x] = ImgDec[y-offset][0];
    }
    for(x=offset; x<width+offset; x++)
    {
      ImgExt[y][x] = ImgDec[y-offset][x-offset];
    }
    for(x=width+offset; x<width+(offset<<1); x++)
    {
      ImgExt[y][x] = ImgDec[y-offset][width-1];
    }
  }
  for(y=height+offset; y<height+(offset<<1); y++)
  {
    for(x=0; x<offset; x++)
    {
      ImgExt[y][x] = ImgDec[height-1][0];
    }
    for(x=offset; x<width+offset; x++)
    {
      ImgExt[y][x] = ImgDec[height-1][x-offset];
    }
    for(x=width+offset; x<width+(offset<<1); x++)
    {
      ImgExt[y][x] = ImgDec[height-1][width-1];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set restored frame
 *****************************************************************************************
 */
void SetRestFrame(imgpel** ImgRest, imgpel** ImgDec, int width, int height)
{
  int i;

  for(i=0; i<height; i++)
    memcpy(ImgDec[i], ImgRest[i], sizeof(imgpel)*width);
}

/*!
 *****************************************************************************************
 * \brief
 *    get base type for filter subtraction
 *****************************************************************************************
 */
int GetALFBaseType(int slice_type, int ref)
{
  int type=0;

  if ( (slice_type==I_SLICE) || (slice_type==SI_SLICE) )
  {
    type = 0;
  }
  else if ( (slice_type==P_SLICE) || (slice_type==SP_SLICE) )
  {
    type = 1;
  }
  else
  {
    type = (ref>0) ? 2 : 3;
  }
  
  return type;
}

/*!
 *****************************************************************************************
 * \brief
 *    initialize base of filter coefficients
 *****************************************************************************************
 */
void InitFilterBase()
{
  int i, tap_idx, type;

  // luma filter base
  for(tap_idx=0; tap_idx<3; tap_idx++)
  {
    for(type=0; type<4; type++)
    {
      for(i=0; i<MAX_NUM_COEF; i++)
      {
        alf_coef_base[type][i] = filter_base_luma[i];
      }
      filter_base_flag[tap_idx][type] = 0;
    }
  }

  // chroma filter base
  for(type=0; type<4; type++)
  {
    for(i=0; i<MAX_NUM_COEF_C; i++)
    {
      alf_coef_base_chroma[type][i] = filter_base_chroma[i];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set difference of filter coefficients 
 *****************************************************************************************
 */
void SetALFDiffCoef(alf_parameter_set_rbsp_t *alfps)
{
  int i, N, tap;
  int *coeff;
  int *filter_pos;
  int *base_flag;
  int **base_coef;

  tap   = alfps->alf.tap;
  N     = (tap*tap+1)>>1;
  coeff = alfps->alf.coeff;

  switch(tap)
  {
  case 5:
    filter_pos = filter_pos_5x5_in_9x9;
    base_flag  = filter_base_flag[0];
    base_coef  = alf_coef_tap_base[0];
    break;
  case 7:
    filter_pos = filter_pos_7x7_in_9x9;
    base_flag  = filter_base_flag[1];
    base_coef  = alf_coef_tap_base[1];
    break;
  case 9:
    filter_pos = filter_pos_9x9_in_9x9;
    base_flag  = filter_base_flag[2];
    base_coef  = alf_coef_tap_base[2];
    break;
  default:
    printf("not support filter size: %d\n", tap);
    exit(1);
    break;
  }

  // set diff coeff
  if(base_flag[base_type] == 0)
  {
    for(i=0; i<N; i++)
    {
      coeff[i] -= alf_coef_base[base_type][filter_pos[i]];
    }
  }
  else
  {
    for(i=0; i<N; i++)
    {
      coeff[i] -= base_coef[base_type][i];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set base of filter coefficients 
 *****************************************************************************************
 */
void SetALFBaseCoef(alf_parameter_set_rbsp_t *alfps)
{
  int i, N, tap;
  int *coeff;
  int *filter_pos;
  int *base_flag;
  int **base_coef;

  tap   = alfps->alf.tap;
  N     = (tap*tap+1)>>1;
  coeff = alfps->alf.coeff;

  switch(tap)
  {
  case 5:
    filter_pos = filter_pos_5x5_in_9x9;
    base_flag  = filter_base_flag[0];
    base_coef  = alf_coef_tap_base[0];
    break;
  case 7:
    filter_pos = filter_pos_7x7_in_9x9;
    base_flag  = filter_base_flag[1];
    base_coef  = alf_coef_tap_base[1];
    break;
  case 9:
    filter_pos = filter_pos_9x9_in_9x9;
    base_flag  = filter_base_flag[2];
    base_coef  = alf_coef_tap_base[2];
    break;
  default:
    printf("not support filter size: %d\n", tap);
    exit(1);
    break;
  }

  // set base coeff
  if(base_flag[base_type] == 0)
  {
    for(i=0; i<N; i++)
    {
      alf_coef_base[base_type][filter_pos[i]] += coeff[i];
      base_coef[base_type][i] = alf_coef_base[base_type][filter_pos[i]];
    }
    base_flag[base_type] = 1;
  }
  else
  {
    for(i=0; i<N; i++)
    {
      base_coef[base_type][i] += coeff[i];
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set difference of filter coefficients for chroma
 *****************************************************************************************
 */
void SetALFDiffCoefChroma(alf_parameter_set_rbsp_t *alfps)
{
  int i, N, tap;

  tap = alfps->alf.tap_chroma;
  N   = (tap*tap+1)>>1;

  for(i=0; i<N; i++)
  {
    alfps->alf.coeff_chroma[i] -= alf_coef_base_chroma[base_type][i];
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set base of filter coefficients for chroma
 *****************************************************************************************
 */
void SetALFBaseCoefChroma(alf_parameter_set_rbsp_t *alfps)
{
  int i, N, tap;

  tap = alfps->alf.tap_chroma;
  N   = (tap*tap+1)>>1;

  for(i=0; i<N; i++)
  {
    alf_coef_base_chroma[base_type][i] += alfps->alf.coeff_chroma[i];
  }
}


/*!
 *****************************************************************************************
 * \brief
 *    calculation lambda for adaptive loop filter
 *****************************************************************************************
 */
void CalcALFLambda(int qp)
{
  if(img->type == B_SLICE && img->nal_reference_idc == 0)
  {
    lambda_alf_luma   = 0.68 * pow (2, (qp - SHIFT_QP) / 3.0) * 3;
    lambda_alf_chroma = lambda_alf_luma;
  }
  else if(img->type == B_SLICE && img->nal_reference_idc == 1)
  {
    lambda_alf_luma   = 0.68 * pow (2, (qp - SHIFT_QP) / 3.0) * 2;
    lambda_alf_chroma = lambda_alf_luma;
  }
  else if(img->type == P_SLICE)
  {
    lambda_alf_luma   = 0.68 * pow (2, (qp - SHIFT_QP) / 3.0);
    lambda_alf_chroma = 0.68 * pow (2, (qp - SHIFT_QP) / 3.0) * 2;
  }
  else
  {
    lambda_alf_luma   = 0.68 * pow (2, (qp - SHIFT_QP) / 3.0);
    lambda_alf_chroma = lambda_alf_luma;
  }
}

/*!
 ************************************************************************
 * \brief
 *    find distortion for luma
 ************************************************************************
 */
int64 FindDistortionFrame(imgpel** ImgOrg, imgpel** ImgCmp)
{
  int i, j;
  int64  diff_y;

  diff_y = 0;
  for (j = 0; j < input->img_height; ++j)
  {
    for (i = 0; i < input->img_width; ++i)
    {
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      diff_y += SQR_DEPTH(ImgOrg[j][i], ImgCmp[j][i], input->BitDepthLuma, img->BitDepthIncrease);
#else
      diff_y += img->quad[ImgOrg[j][i] - ImgCmp[j][i]];
#endif
    }
  }

  return diff_y;
}

/*!
 ************************************************************************
 * \brief
 *    find distortion for chroma
 ************************************************************************
 */
void FindDistortionFrameChroma(imgpel*** ImgUVOrg, imgpel*** ImgUVCmp, int64 *dist_u, int64 *dist_v)
{
  int i, j;
  int64 diff_u, diff_v;

  diff_u = diff_v = 0;
  if (img->yuv_format != YUV400)
  {
    for (j = 0; j < input->img_height_cr; j++)
    {
      for (i = 0; i < input->img_width_cr; i++)
      {
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        diff_u += SQR_DEPTH(ImgUVOrg[0][j][i], ImgUVCmp[0][j][i], input->BitDepthChroma, img->BitDepthIncreaseChroma);
        diff_v += SQR_DEPTH(ImgUVOrg[1][j][i], ImgUVCmp[1][j][i], input->BitDepthChroma, img->BitDepthIncreaseChroma);
#else
        diff_u += img->quad[ImgUVOrg[0][j][i] - ImgUVCmp[0][j][i]];
        diff_v += img->quad[ImgUVOrg[1][j][i] - ImgUVCmp[1][j][i]];
#endif
      }
    }
  }

  *dist_u = diff_u;
  *dist_v = diff_v;
}

/*!
 ************************************************************************
 * \brief
 *    find distortion for block
 ************************************************************************
 */
int64 FindDistortionBlock(int blk_addr, imgpel** ImgOrg, imgpel** ImgCmp, int blk_size, int width_in_alf_blks)
{
  int x, y, blk_x, blk_y, max_blk_x, max_blk_y;
  int64  diff_y;

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, width_in_alf_blks);

  max_blk_x = blk_x + blk_size;
  max_blk_y = blk_y + blk_size;
  if(max_blk_x > img->width)  max_blk_x = img->width;
  if(max_blk_y > img->height) max_blk_y = img->height;

  diff_y = 0;
  for (y = blk_y; y < max_blk_y; y++)
  {
    for (x = blk_x; x < max_blk_x; x++)
    {
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      diff_y += SQR_DEPTH(ImgOrg[y][x], ImgCmp[y][x], input->BitDepthLuma, img->BitDepthIncrease);
#else
      diff_y += img->quad[ImgOrg[y][x] - ImgCmp[y][x]];
#endif
    }
  }

  return diff_y;
}

/*!
 *****************************************************************************************
 * \brief
 *    calculation Rate-Distortion cost
 *****************************************************************************************
 */
void CalcRDCost(imgpel** ImgOrg, imgpel** ImgCmp, alf_parameter_set_rbsp_t *alfps, int64* rate, int64* dist, double* cost)
{
  int *tmp_coef;

  if(alfps != NULL)
  {
    get_mem1Dint    (&(tmp_coef), MAX_NUM_COEF);
    memcpy(tmp_coef, alfps->alf.coeff, sizeof(int)*alfps->alf.num_coeff);
    
    SetALFDiffCoef(alfps);

    *rate      = send_ALFParameterSet(alfps, tempStream);

    memcpy(alfps->alf.coeff, tmp_coef, sizeof(int)*alfps->alf.num_coeff);
    free_mem1Dint(tmp_coef);
  }
  else
  {
    *rate    = 0;
  }
  *dist      = FindDistortionFrame(ImgOrg, ImgCmp);
  *cost      = (double)(*rate) * lambda_alf_luma + (double)(*dist);
}

/*!
 *****************************************************************************************
 * \brief
 *    calculation Rate-Distortion cost for Chroma
 *****************************************************************************************
 */
void CalcRDCostChroma(imgpel*** ImgUVOrg, imgpel*** ImgUVCmp, alf_parameter_set_rbsp_t *alfps, int64* rate, int64* dist, double* cost)
{
  int *tmp_coef;
  int64 dist_u, dist_v;

  if(alfps->chroma_idc)
  {
    get_mem1Dint    (&(tmp_coef), MAX_NUM_COEF_C);
    memcpy(tmp_coef, alfps->alf.coeff_chroma, sizeof(int)*alfps->alf.num_coeff_chroma);
    
    SetALFDiffCoefChroma(alfps);

    *rate      = send_ALFParameterSet(alfps, tempStream);

    memcpy(alfps->alf.coeff_chroma, tmp_coef, sizeof(int)*alfps->alf.num_coeff_chroma);
    free_mem1Dint(tmp_coef);
  }
  else
  {
    *rate    = tmp_luma_rate;
  }
  FindDistortionFrameChroma(ImgUVOrg, ImgUVCmp, &dist_u, &dist_v);
  *dist      = dist_u + dist_v;
  *cost      = (double)(*rate) * lambda_alf_chroma + (double)(*dist);
}

/*!
 *****************************************************************************************
 * \brief
 *    calculation Rate-Distortion cost for Block-based ALF
 *****************************************************************************************
 */
void CalcRDCostBlock(imgpel** ImgOrg, imgpel** ImgCmp, imgpel** ImgDec, alf_parameter_set_rbsp_t *alfps, int64* rate, int64* dist, double* cost)
{
  int i;
  int *tmp_coef;
  AlfBlock *currAlfBlk;

  if(alfps != NULL)
  {
    get_mem1Dint    (&(tmp_coef), MAX_NUM_COEF);
    memcpy(tmp_coef, alfps->alf.coeff, sizeof(int)*alfps->alf.num_coeff);

    SetALFDiffCoef(alfps);

    *rate      = send_ALFParameterSet(alfps, tempStream);

    memcpy(alfps->alf.coeff, tmp_coef, sizeof(int)*alfps->alf.num_coeff);
    free_mem1Dint(tmp_coef);
  }
  else
  {
    *rate   = 0;
  }
  *dist     = 0;
  for (i=0; i<alfps->num_alf_block; i++)
  {
    currAlfBlk = &alfps->alf_blk_data[i];
    if(currAlfBlk->alf_blk_flag)
    {
      *dist += FindDistortionBlock(i, ImgOrg, ImgCmp, alfps->alf_block_size, alfps->WidthInAlfBlks);
    }
    else
    {
      *dist += FindDistortionBlock(i, ImgOrg, ImgDec, alfps->alf_block_size, alfps->WidthInAlfBlks);
    }
  }
  *cost     = (double)(*rate) * lambda_alf_luma + (double)(*dist);
}


/*!
 *****************************************************************************************
 * \brief
 *    calculates correlation function
 *****************************************************************************************
 */
void CalcCorrelationFunc(imgpel** ImgOrg, imgpel** ImgDec, double** corr, int tap, int width, int height)
{
  int i, j, x, y, N, offset;
  imgpel *term;
  int *pFiltPos;
  int yy, xx;

  N      = (tap*tap+1)>>1;
  offset = tap>>1;

  switch(tap)
  {
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  if((term = (imgpel *)calloc(N, sizeof(imgpel))) == NULL)
    no_mem_exit("CalcCorrelationFunc: term");

  for(y=ext_offset; y<height+ext_offset; y++)
  {
    for(x=ext_offset; x<width+ext_offset; x++)
    {
      i=0;
      memset(term, 0, sizeof(imgpel)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          term[pFiltPos[i]] += ImgDec[yy][xx];
          i++;
        }
      }

      for(j=0; j<N; j++)
      {
        corr[j][j] += term[j]*term[j];
        for(i=j+1; i<N; i++)
          corr[j][i] += term[j]*term[i];

        // DC offset
        corr[j][N]   += term[j];
        corr[j][N+1] += ImgOrg[y-ext_offset][x-ext_offset]*term[j];
      }
      // DC offset
      for(i=0; i<N; i++)
        corr[N][i] += term[i];
      corr[N][N]   += 1;
      corr[N][N+1] += ImgOrg[y-ext_offset][x-ext_offset];
    }
  }
  for(j=0; j<N-1; j++)
    for(i=j+1; i<N; i++)
      corr[i][j] = corr[j][i];

  free(term);
}

/*!
 *****************************************************************************************
 * \brief
 *    Calculates correlation function for Block-based ALF
 *****************************************************************************************
 */
void CalcCorrelationFuncBlock(int blk_addr, imgpel** ImgOrg, imgpel** ImgDec, unsigned int** corr, int tap)
{
  int i, j, x, y, N, offset;
  int blk_size, blk_x, blk_y, max_blk_x, max_blk_y;
  unsigned int *term;
  int *pFiltPos;
  int yy, xx;

  N        = (tap*tap+1)>>1;
  offset   = tap>>1;
  blk_size = alf_blk_size[0];

  switch(tap)
  {
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  if((term = (unsigned int *)calloc(N, sizeof(unsigned int))) == NULL)
    no_mem_exit("CalcCorrelationFuncBlock: term");

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, img->max_WidthInAlfBlks);

  max_blk_x = blk_x + blk_size;
  max_blk_y = blk_y + blk_size;
  if(max_blk_x > img->width)  max_blk_x = img->width;
  if(max_blk_y > img->height) max_blk_y = img->height;

  for(y=blk_y+ext_offset; y<max_blk_y+ext_offset; y++)
  {
    for(x=blk_x+ext_offset; x<max_blk_x+ext_offset; x++)
    {
      i=0;
      memset(term, 0, sizeof(unsigned int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          term[pFiltPos[i]] += (unsigned int)ImgDec[yy][xx];
          i++;
        }
      }

      for(j=0; j<N; j++)
      {
        corr[j][0] += term[j]*term[j];
        for(i=j+1; i<N; i++)
          corr[j][i-j] += term[j]*term[i];

        // DC offset
        corr[j][N-j]   += term[j];
        corr[j][N-j+1] += (unsigned int)ImgOrg[y-ext_offset][x-ext_offset]*term[j];
      }
      // DC offset
      corr[N][0] += 1;
      corr[N][1] += (unsigned int)ImgOrg[y-ext_offset][x-ext_offset];
    }
  }

  free(term);
}

/*!
 *****************************************************************************************
 * \brief
 *    calculates correlation function (5x5->7x7, 7x7->9x9)
 *****************************************************************************************
 */
void CalcCorrelationFuncBlockDiff(int blk_addr, imgpel** ImgOrg, imgpel** ImgDec, unsigned int** corr, int tap)
{
  int i, j, x, y, posi, posj, N, offset;
  int blk_size, blk_x, blk_y, max_blk_x, max_blk_y;
  unsigned int *term;
  int *diff, *same;
  int num_diff;
  int end_same;
  int start_diff;
  int *pFiltPos;
  int yy, xx;

  N        = (tap*tap+1)>>1;
  offset   = tap>>1;
  blk_size = alf_blk_size[0];

  switch(tap)
  {
  case 7:
    num_diff = NUM_COEF_7x7-NUM_COEF_5x5;
    end_same = NUM_COEF_5x5-(offset+1);
    diff     = diff_pos_5x5_in_7x7;
    same     = same_pos_5x5_in_7x7;
    pFiltPos = symmetric_array7x7;
    break;
  case 9:
    num_diff = MAX_NUM_COEF-NUM_COEF_7x7;
    end_same = NUM_COEF_7x7-(offset+1);
    diff     = diff_pos_7x7_in_9x9;
    same     = same_pos_7x7_in_9x9;
    pFiltPos = symmetric_array9x9;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  if((term = (unsigned int *)calloc(N, sizeof(unsigned int))) == NULL)
    no_mem_exit("CalcCorrelationFuncBlockDiff: term");

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, img->max_WidthInAlfBlks);

  max_blk_x = blk_x + blk_size;
  max_blk_y = blk_y + blk_size;
  if(max_blk_x > img->width)  max_blk_x = img->width;
  if(max_blk_y > img->height) max_blk_y = img->height;

  for(y=blk_y+ext_offset; y<max_blk_y+ext_offset; y++)
  {
    for(x=blk_x+ext_offset; x<max_blk_x+ext_offset; x++)
    {
      i=0;
      memset(term, 0, sizeof(unsigned int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          term[pFiltPos[i]] += (unsigned int)ImgDec[yy][xx];
          i++;
        }
      }

      for(j=0; j<num_diff; j++)
      {
        posj = diff[j];
        corr[posj][0] += term[posj]*term[posj];
        for(i=posj+1; i<N; i++)
          corr[posj][i-posj] += term[posj]*term[i];

        // DC offset
        corr[posj][N-posj]   += term[posj];
        corr[posj][N+1-posj] += (unsigned int)ImgOrg[y-ext_offset][x-ext_offset]*term[posj];
      }
      for(j=0; j<end_same; j++)
      {
        posj = same[j];
        start_diff = (tap+1)+((j/(tap-2))<<1);
        for(i=start_diff; i<num_diff; i++)
        {
          posi = diff[i];
          corr[posj][posi-posj] += term[posj]*term[posi];
        }
      }
    }
  }

  free(term);
}

/*!
 *****************************************************************************************
 * \brief
 *    copy correlation
 *****************************************************************************************
 */
void CopyBlockCorrelation(unsigned int **source, unsigned int **dest, int tap_idx)
{
  int i, j;
  int posi, posj;
  int sourceN, destN;
  int *filter_pos;

  if(tap_idx==0)
  {
    sourceN    = 13;
    destN      = 25;
    filter_pos = filter_pos_5x5_in_7x7;
  }
  else
  {
    sourceN    = 25;
    destN      = 41;
    filter_pos = filter_pos_7x7_in_9x9;
  }

  for(j=0; j<sourceN; j++)
  {
    posj = filter_pos[j];
    for(i=j; i<sourceN; i++)
    {
      posi = filter_pos[i];
      dest[posj][posi-posj] = source[j][i-j];
    }
    dest[posj][destN-posj]   = source[j][sourceN-j];

    // DC offset
    dest[posj][destN+1-posj] = source[j][sourceN+1-j];
  }
  // DC offset
  dest[destN][0] = source[sourceN][0];
  dest[destN][1] = source[sourceN][1];
}

/*!
 *****************************************************************************************
 * \brief
 *    calcurates correlation function for variable filter-tap
 *****************************************************************************************
 */
void CalcCorrelationFuncVariableTap(imgpel** ImgOrg, imgpel** ImgDec)
{
  int i, j, k, bx, by;
  unsigned int  **p_blk_corr[3];

  for (i=0; i<img->max_num_alf_block; i++)
  {
    by = i / img->max_WidthInAlfBlks;
    bx = i % img->max_WidthInAlfBlks;

    p_blk_corr[0] = blk_corr5x5[by][bx];
    p_blk_corr[1] = blk_corr7x7[by][bx];
    p_blk_corr[2] = blk_corr[by][bx];

    for(j=0; j<NUM_COEF_5x5; j++)
      memset(p_blk_corr[0][j], 0, sizeof(unsigned int)*(NUM_COEF_5x5+1-j));
    for(j=0; j<NUM_COEF_7x7; j++)
      memset(p_blk_corr[1][j], 0, sizeof(unsigned int)*(NUM_COEF_7x7+1-j));
    for(j=0; j<MAX_NUM_COEF; j++)
      memset(p_blk_corr[2][j], 0, sizeof(unsigned int)*(MAX_NUM_COEF+1-j));
    if(i==0)
    {
      for(j=0; j<NUM_COEF_5x5; j++)
        memset(correlation[0][j], 0, sizeof(double)*(NUM_COEF_5x5+1));
      for(j=0; j<NUM_COEF_7x7; j++)
        memset(correlation[1][j], 0, sizeof(double)*(NUM_COEF_7x7+1));
      for(j=0; j<MAX_NUM_COEF; j++)
        memset(correlation[2][j], 0, sizeof(double)*(MAX_NUM_COEF+1));
    }

    CalcCorrelationFuncBlock(i, ImgOrg, ImgDec, p_blk_corr[0], 5);
    CopyBlockCorrelation(p_blk_corr[0], p_blk_corr[1], 0);
    CalcCorrelationFuncBlockDiff(i, ImgOrg, ImgDec, p_blk_corr[1], 7);
    CopyBlockCorrelation(p_blk_corr[1], p_blk_corr[2], 1);
    CalcCorrelationFuncBlockDiff(i, ImgOrg, ImgDec, p_blk_corr[2], 9);

    for(j=0; j<NUM_COEF_5x5; j++)
      for(k=j; k<NUM_COEF_5x5+1; k++)
        correlation[0][j][k] += (double)p_blk_corr[0][j][k-j];
    for(j=0; j<NUM_COEF_7x7; j++)
      for(k=j; k<NUM_COEF_7x7+1; k++)
        correlation[1][j][k] += (double)p_blk_corr[1][j][k-j];
    for(j=0; j<MAX_NUM_COEF; j++)
      for(k=j; k<MAX_NUM_COEF+1; k++)
        correlation[2][j][k] += (double)p_blk_corr[2][j][k-j];
  }
  for(j=0; j<NUM_COEF_5x5-1; j++)
    for(k=j+1; k<NUM_COEF_5x5; k++)
      correlation[0][k][j] = correlation[0][j][k];
  for(j=0; j<NUM_COEF_7x7-1; j++)
    for(k=j+1; k<NUM_COEF_7x7; k++)
      correlation[1][k][j] = correlation[1][j][k];
  for(j=0; j<MAX_NUM_COEF-1; j++)
    for(k=j+1; k<MAX_NUM_COEF; k++)
      correlation[2][k][j] = correlation[2][j][k];
}

/*!
 *****************************************************************************************
 * \brief
 *    gaussian elimination
 *****************************************************************************************
 */
int Gauss(double **a, int N)
{
  int i, j, k;
  double t;

  for(k=0; k<N-1; k++)
  {
    for(i=k+1;i<N; i++)
    {
      t=a[i][k]/a[k][k];
      for(j=k+1; j<=N; j++)
      {
        a[i][j] -= t * a[k][j];
        if(i==j && fabs(a[i][j])<0.000001) return 1;
      }
    }
  }
  for(i=N-1; i>=0; i--)
  {
    t = a[i][N];
    for(j=i+1; j<N; j++)
      t -= a[i][j] * a[j][N];
    a[i][N] = t / a[i][i];
  }
  return 0;
}

/*!
 ************************************************************************
 * \brief
 *    quick sort filter coefficients
 ************************************************************************
 */
void FilterCoefQuickSort( double *coef_data, int *coef_num, int upper, int lower )
{
  double mid, tmp_data;
  int i, j, tmp_num;

  i = upper;
  j = lower;
  mid = coef_data[(lower+upper)>>1];
  do
  {
    while( coef_data[i] < mid ) i++;
    while( mid < coef_data[j] ) j--;
    if( i <= j )
    {
      tmp_data = coef_data[i];
      tmp_num  = coef_num[i];
      coef_data[i] = coef_data[j];
      coef_num[i]  = coef_num[j];
      coef_data[j] = tmp_data;
      coef_num[j]  = tmp_num;
      i++;
      j--;
    }
  } while( i <= j );
  if ( upper < j ) FilterCoefQuickSort(coef_data, coef_num, upper, j);
  if ( i < lower ) FilterCoefQuickSort(coef_data, coef_num, i, lower);
}

/*!
 *****************************************************************************************
 * \brief
 *    type transition from double to int and quantization 
 *****************************************************************************************
 */
void QuantFilterCoef(double* h, int* qh, int tap, int bit_depth)
{
  int i, N;
  int max_value, min_value;
  double dbl_total_gain;
  int total_gain, q_total_gain;
  int upper, lower;
  double *dh;
  int    *nc;
  int    *pFiltMag;

  switch(tap)
  {
  case 5:
    pFiltMag = symmetric_mag5x5;
    break;
  case 7:
    pFiltMag = symmetric_mag7x7;
    break;
  case 9:
    pFiltMag = symmetric_mag9x9;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  N = (tap*tap+1)>>1;

  get_mem1Ddouble(&(dh), N);
  get_mem1Dint(&(nc), N);

  max_value =   (1<<(1+num_bit))-1;
  min_value = 0-(1<<(1+num_bit));

  dbl_total_gain=0.0;
  q_total_gain=0;
  for(i=0; i<N; i++)
  {
    if(h[i]>=0.0)
      qh[i] =  (int)( h[i]*(1<<num_bit)+0.5);
    else
      qh[i] = -(int)(-h[i]*(1<<num_bit)+0.5);

    dh[i] = (double)qh[i]/(double)(1<<num_bit) - h[i];
    dh[i]*=pFiltMag[i];
    dbl_total_gain += h[i]*pFiltMag[i];
    q_total_gain   += qh[i]*pFiltMag[i];
    nc[i] = i;
  }

  // modification of quantized filter coefficients
  total_gain = (int)(dbl_total_gain*(1<<num_bit)+0.5);

  if( q_total_gain != total_gain )
  {
    FilterCoefQuickSort(dh, nc, 0, N-1);
    if( q_total_gain > total_gain )
    {
      upper = N-1;
      while( q_total_gain > total_gain+1 )
      {
        i = nc[upper%N];
        qh[i]--;
        q_total_gain -= pFiltMag[i];
        upper--;
      }
      if( q_total_gain == total_gain+1 )
      {
        if(dh[N-1]>0)
          qh[N-1]--;
        else
        {
          i=nc[upper%N];
          qh[i]--;
          qh[N-1]++;
        }
      }
    }
    else if( q_total_gain < total_gain )
    {
      lower = 0;
      while( q_total_gain < total_gain-1 )
      {
        i=nc[lower%N];
        qh[i]++;
        q_total_gain += pFiltMag[i];
        lower++;
      }
      if( q_total_gain == total_gain-1 )
      {
        if(dh[N-1]<0)
          qh[N-1]++;
        else
        {
          i=nc[lower%N];
          qh[i]++;
          qh[N-1]--;
        }
      }
    }
  }
  
  // set of filter coefficients
  for(i=0; i<N; i++)
  {
    if(qh[i]<min_value || qh[i]>max_value) printf("qh[%d] = %d \n", i, qh[i]);
    qh[i] = max(min_value,min(max_value, qh[i]));
  }
  // DC offset
  max_value = min(  (1<<(3+max(img->bitdepth_luma,img->bitdepth_chroma)))-1, (1<<14)-1);
  min_value = max( -(1<<(3+max(img->bitdepth_luma,img->bitdepth_chroma))),  -(1<<14)  );
  qh[N] =  (h[N]>=0.0)? (int)( h[N]*(1<<(num_bit-bit_depth+8)) + 0.5) : -(int)(-h[N]*(1<<(num_bit-bit_depth+8)) + 0.5);
  qh[N] = max(min_value,min(max_value, qh[N]));

  free_mem1Ddouble(dh);
  free_mem1Dint(nc);
}

/*!
 *****************************************************************************************
 * \brief
 *    clear filter coefficients
 *****************************************************************************************
 */
void ClearFilterCoefInt(int* qh, int N)
{
  int i;

  for(i=0; i<N; i++)
  {
    qh[i] = 0;
  }
  // center pos
  qh[N-2]  = 1<<num_bit;
}


//------- filtering for luma
/*!
 *****************************************************************************************
 * \brief
 *    adaptive loop filter main process
 *****************************************************************************************
 */
void AdaptiveLoopFilterProcess(Picture *pic)
{
  int tap, num_coef;

  Slice *currSlice;
  imgpel **ImgOrg;
  imgpel **ImgDec;
  imgpel **ImgRest;
  
  // set global variables
  tap         = MAX_NUM_TAP;
  num_coef    = (tap*tap+1)>>1;
  num_coef    = num_coef + 1; // DC offset
  num_bit     = NUM_BIT_SHIFT;
  r_offset    = 1<<(num_bit-1);
  ext_offset  = tap>>1;
  base_type   = GetALFBaseType(img->type, img->nal_reference_idc);

  img->currentPicture = pic;
  img->currentSlice   = pic->slices[0];
  img->qp             = pic->slices[0]->qp;

  if(input->symbol_mode == CABAC)
    img->model_number   = pic->slices[0]->model_number;

  tempStream = NULL;

  // initialize base coefficients
  if(img->currentPicture->idr_flag)
  {
    InitFilterBase();
  }

  // set lambda
  CalcALFLambda(img->qp);

  // extend image for filtering
  MakeExtendFrame( enc_picture->imgY, imgY_ext, img->width, img->height, ext_offset );

  // set image buffer pointer
  ImgOrg = imgY_org;
  ImgDec = imgY_ext;
  ImgRest= imgY_rest;

  // set min cost
  min_rate = 0x7fffffffffffffffLL;
  min_dist = 0x7fffffffffffffffLL;
  min_cost = DBL_MAX;

  // calc original cost
  CalcRDCost(ImgOrg, enc_picture->imgY, NULL, &orig_rate, &orig_dist, &orig_cost);

  // initialize temp_alfps
  temp_alfps->alf_flag      = 1;
  temp_alfps->alf.tap       = tap;
  temp_alfps->alf.num_coeff = num_coef;
  temp_alfps->chroma_idc    = 0;
  temp_alfps->block_control_flag = 0;

  // Adaptive in-loop wiener filtering
  FilterTapDecision(ImgOrg, ImgDec, ImgRest);

  // block-based filter on/off control
  BlockAdaptiveFilterControl(ImgOrg, ImgDec, ImgRest);

  CalcRDCost(ImgOrg, ImgRest, active_alfps, &rate, &dist, &cost);

  if( cost < orig_cost )
  {
    active_alfps->alf_flag = 1;
  }
  else
  {
    active_alfps->alf_flag = 0;
  }

  if(active_alfps->alf_flag)
  {
    // set alf_flag in slice header
    currSlice = img->currentSlice;
    *(currSlice->alf_flag_stream_pointer) += 1<<(currSlice->alf_flag_bit_to_go-1);

    // set filtered frame
    SetRestFrame(ImgRest, enc_picture->imgY, img->width, img->height);

    // set diff coefficients
    SetALFDiffCoef(active_alfps);

    // set base coefficients
    SetALFBaseCoef(active_alfps);

    // Adaptive in-loop wiener filtering for chroma
    tmp_luma_rate = rate;
    AdaptiveLoopFilterProcessChroma();

    // make bitstream for ALF
    send_ALFParameterSet(active_alfps, alfstream);
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    decision of filter-tap by Rate-Distortion Optimization
 *****************************************************************************************
 */
void FilterTapDecision(imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest)
{
  int tap, num_coef;

  // calc. correlation for variable tap
  CalcCorrelationFuncVariableTap(ImgOrg, ImgDec);

  for(tap=MIN_NUM_TAP; tap<=MAX_NUM_TAP; tap+=2)
  {
    // restriction for non-stored B-slice
    if(img->type==B_SLICE && img->nal_reference_idc == 0)
    {
      if(tap > 5) continue;
    }

    temp_alfps->alf.tap = tap;
    num_coef            = (tap*tap+1)>>1;
    num_coef            = num_coef + 1; // DC offset
    temp_alfps->alf.num_coeff = num_coef;

    FilteringFrame(ImgDec, imgY_temp);

    CalcRDCost(ImgOrg, imgY_temp, temp_alfps, &rate, &dist, &cost);

    if( cost < min_cost )
    {
      min_rate = rate;
      min_dist = dist;
      min_cost = cost;
      SetALFParameters(active_alfps, temp_alfps);
      SetRestFrame(imgY_temp, ImgRest, img->width, img->height);
    }
  }

  // set temp_alfps
  SetALFParameters(temp_alfps, active_alfps);
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering frame for luma
 *****************************************************************************************
 */
void FilteringFrame(imgpel** ImgDec, imgpel** ImgRest)
{
  int i, tap, N, err_code;
  double **corr, *h;
  int    *qh;

  tap  = temp_alfps->alf.tap;
  N    = temp_alfps->alf.num_coeff;
  h    = dbl_alf_coef;
  qh   = temp_alfps->alf.coeff;

  switch(tap)
  {
  case 5:
    corr = correlation[0];
    break;
  case 7:
    corr = correlation[1];
    break;
  case 9:
    corr = correlation[2];
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }
    
  err_code = Gauss(corr, N);
  if(err_code)
  {
    ClearFilterCoefInt(qh, N);
  }
  else
  {
    for(i=0; i<N; i++)
      h[i] = corr[i][N];

    QuantFilterCoef(h, qh, tap, img->bitdepth_luma);
  }

  FilteringProcess(ImgDec, ImgRest, qh, tap);
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering process for luma
 *****************************************************************************************
 */
void FilteringProcess(imgpel** ImgDec, imgpel** ImgRest, int *qh, int tap)
{
  int i, x, y, value, N, offset;
  int *pFiltPos, PixSum[MAX_NUM_COEF];
  int yy, xx;

  N      = (tap*tap+1)>>1;
  offset = tap>>1;

  switch(tap)
  {
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  for(y=ext_offset; y<img->height+ext_offset; y++)
  {
    for(x=ext_offset; x<img->width+ext_offset; x++)
    {
      i=0;
      memset(PixSum, 0, sizeof(int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          PixSum[pFiltPos[i]] += (int)ImgDec[yy][xx];
          i++;
        }
      }
      value = 0;
      for(i=0; i<N; i++)
      {
        value += qh[i]*PixSum[i];
      }
      // DC offset
      value += qh[N]<<(img->bitdepth_luma-8);
      value = (value + r_offset)>>num_bit;
      ImgRest[y-ext_offset][x-ext_offset] = (imgpel)Clip1(value);
    }
  }
}


//------- filtering for chroma
/*!
 *****************************************************************************************
 * \brief
 *    adaptive loop filter main process for chroma
 *****************************************************************************************
 */
void AdaptiveLoopFilterProcessChroma()
{
  int tap, num_coef;
  int64 dist_u, dist_v, dist_u_rest, dist_v_rest;

  imgpel ***ImgUVOrg;
  imgpel ***ImgUVDec;
  imgpel ***ImgUVRest;

  // restriction for non-stored B-slice
  if(img->nal_reference_idc == 0)
  {
    active_alfps->chroma_idc = 0;
    return;
  }

  // set global variables for chroma
  tap        = MAX_NUM_TAP_C;
  ext_offset = tap>>1;
  num_coef   = (tap*tap+1)>>1;
  num_coef   = num_coef + 1;  // DC offset
  base_type  = GetALFBaseType(img->type, img->nal_reference_idc);

  // extend picture for filtering
  MakeExtendFrame( enc_picture->imgUV[0], imgUV_ext[0], img->width_cr, img->height_cr, ext_offset );
  MakeExtendFrame( enc_picture->imgUV[1], imgUV_ext[1], img->width_cr, img->height_cr, ext_offset );
  
  // set image buffer pointer
  ImgUVOrg  = imgUV_org;
  ImgUVDec  = imgUV_ext;
  ImgUVRest = imgUV_rest;

  // calc. original cost
  SetALFParameters(temp_alfps, active_alfps);
  CalcRDCostChroma(ImgUVOrg, enc_picture->imgUV, temp_alfps, &min_rate, &min_dist, &min_cost);

  // initialize temp_alfps
  temp_alfps->chroma_idc           = 3;
  temp_alfps->alf.tap_chroma       = tap;
  temp_alfps->alf.num_coeff_chroma = num_coef;

  // Adaptive in-loop wiener filtering for chroma
  FilteringFrameChroma(ImgUVOrg, ImgUVDec, ImgUVRest);

  // filter on/off decision for chroma
  FindDistortionFrameChroma( ImgUVOrg, enc_picture->imgUV, &dist_u,      &dist_v      );
  FindDistortionFrameChroma( ImgUVOrg, ImgUVRest,          &dist_u_rest, &dist_v_rest );

  temp_alfps->chroma_idc = 0;
  if(dist_u > dist_u_rest)
    temp_alfps->chroma_idc += 2;
  if(dist_v  > dist_v_rest )
    temp_alfps->chroma_idc += 1;

  if(temp_alfps->chroma_idc)
  {
    if(temp_alfps->chroma_idc!=3)
    {
      // chroma filter re-design
      FilteringFrameChroma(ImgUVOrg, ImgUVDec, ImgUVRest);
    }

    CalcRDCostChroma(ImgUVOrg, ImgUVRest, temp_alfps, &rate, &dist, &cost);

    if( cost < min_cost )
    {
      SetALFParameters(active_alfps, temp_alfps);

      // set filtered frame
      if((active_alfps->chroma_idc>>1)&0x01)
        SetRestFrame(ImgUVRest[0], enc_picture->imgUV[0], img->width_cr, img->height_cr);
      if(active_alfps->chroma_idc&0x01)
        SetRestFrame(ImgUVRest[1], enc_picture->imgUV[1], img->width_cr, img->height_cr);

      // set diff coefficients
      SetALFDiffCoefChroma(active_alfps);

      // set base coefficients
      SetALFBaseCoefChroma(active_alfps);
    }
    else
    {
      active_alfps->chroma_idc = 0;
    }
  }
  else
  {
    active_alfps->chroma_idc = 0;
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering frame for chroma
 *****************************************************************************************
 */
void FilteringFrameChroma(imgpel ***ImgUVOrg, imgpel ***ImgUVDec, imgpel ***ImgUVRest)
{
  int i, tap, N, err_code;
  double **corr, *h;
  int    *qh;

  tap  = temp_alfps->alf.tap_chroma;
  N    = temp_alfps->alf.num_coeff_chroma;
  corr = correlation[0];
  h    = dbl_alf_coef;
  qh   = temp_alfps->alf.coeff_chroma;

  // initialize correlation
  for(i=0; i<N; i++)
    memset(corr[i], 0, sizeof(double)*(N+1));

  if((temp_alfps->chroma_idc>>1)&0x01)
  {
    CalcCorrelationFunc(ImgUVOrg[0], ImgUVDec[0], corr, tap, img->width_cr, img->height_cr);
  }
  if((temp_alfps->chroma_idc)&0x01)
  {
    CalcCorrelationFunc(ImgUVOrg[1], ImgUVDec[1], corr, tap, img->width_cr, img->height_cr);
  }

  err_code = Gauss(corr, N);
  if(err_code)
  {
    ClearFilterCoefInt(qh, N);
  }
  else
  {
    for(i=0; i<N; i++)
      h[i] = corr[i][N];

    QuantFilterCoef(h, qh, tap, img->bitdepth_chroma);
  }

  if((temp_alfps->chroma_idc>>1)&0x01)
  {
    FilteringProcessChroma(ImgUVDec, ImgUVRest, qh, tap, 0);
  }
  if((temp_alfps->chroma_idc)&0x01)
  {
    FilteringProcessChroma(ImgUVDec, ImgUVRest, qh, tap, 1);
  }

  if(temp_alfps->chroma_idc<3)
  {
    if(temp_alfps->chroma_idc==1)
    {
      SetRestFrame(enc_picture->imgUV[0], ImgUVRest[0], img->width_cr, img->height_cr);
    }
    if(temp_alfps->chroma_idc==2)
    {
      SetRestFrame(enc_picture->imgUV[1], ImgUVRest[1], img->width_cr, img->height_cr);
    }
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    filtering process for chroma
 *****************************************************************************************
 */
void FilteringProcessChroma(imgpel ***ImgUVDec, imgpel ***ImgUVRest, int *qh, int tap, int chroma_id)
{
  int i, x, y, value, N, offset;
  int *pFiltPos, PixSum[MAX_NUM_COEF_C];
  int yy, xx;

  N      = (tap*tap+1)>>1;
  offset = tap>>1;

  switch(tap)
  {
  case 5:
    pFiltPos = symmetric_array5x5;
    break;
  case 7:
    pFiltPos = symmetric_array7x7;
    break;
  case 9:
    pFiltPos = symmetric_array9x9;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  for(y=ext_offset; y<img->height_cr+ext_offset; y++)
  {
    for(x=ext_offset; x<img->width_cr+ext_offset; x++)
    {
      i=0;
      memset(PixSum, 0, sizeof(int)*N);
      for(yy=y-offset; yy<=y+offset; yy++)
      {
        for(xx=x-offset; xx<=x+offset; xx++)
        {
          PixSum[pFiltPos[i]] += (int)ImgUVDec[chroma_id][yy][xx];
          i++;
        }
      }
      value = 0;
      for(i=0; i<N; i++)
      {
        value += qh[i]*PixSum[i];
      }
      // DC offset
      value += qh[N]<<(img->bitdepth_chroma-8);

      value = (value + r_offset)>>num_bit;
      ImgUVRest[chroma_id][y-ext_offset][x-ext_offset] = (imgpel)Clip1_Chr(value);
    }
  }
}


//------- filtering based block
/*!
 *****************************************************************************************
 * \brief
 *    block-based adaptive loop filtering
 *****************************************************************************************
 */
void BlockAdaptiveFilterControl(imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest)
{
  int i, idx, not_balf_flag;
  int size, wnum, hnum, tnum;
  AlfBlock *currAlfBlk;

  temp_alfps->block_control_flag = 1;

  // block size decision by Rate-Distortion Optimization
  for(idx=0; idx<8; idx++)
  {
    // block size restrictions
    size = alf_blk_size[idx];
    if(size==24 || size==48 || size==96) continue;
    if((img->qp>=30)&&(size==8)) continue;
    wnum = img->width  / alf_blk_size[idx];
    hnum = img->height / alf_blk_size[idx];
    if( (img->width  % alf_blk_size[idx]) != 0 ) wnum++;
    if( (img->height % alf_blk_size[idx]) != 0 ) hnum++;
    tnum = wnum * hnum;
    if( tnum <= 30 ) continue;
    if( (tnum>=3000) && (img->type==B_SLICE) ) continue;

    // set block size parameter
    SetBlockSizeParameter(temp_alfps, idx);

    // set block-based on/off control flag
    not_balf_flag = SetALFBlockFlag(temp_alfps, ImgOrg, enc_picture->imgY, ImgRest);
    if(not_balf_flag) continue;

    CalcRDCostBlock(ImgOrg, ImgRest, enc_picture->imgY, temp_alfps, &rate, &dist, &cost);

    if( cost < min_cost )
    {
      min_rate = rate;
      min_dist = dist;
      min_cost = cost;
      SetALFParameters(active_alfps, temp_alfps);
      SetRestFrame(ImgRest, imgY_best, img->width, img->height);
    }

    // re-design filter coefficients
    ReDesignFilterDecision(ImgOrg, ImgDec);
  }

  if(active_alfps->block_control_flag)
  {
    for(i=0; i<active_alfps->num_alf_block; i++)
    {
      currAlfBlk = &active_alfps->alf_blk_data[i];

      if(!currAlfBlk->alf_blk_flag)
      {
        CopyDecToRestBlock(i, enc_picture->imgY, imgY_best, active_alfps->alf_block_size, active_alfps->WidthInAlfBlks);
      }
    }
    SetRestFrame(imgY_best, ImgRest, img->width, img->height);
    SetBlockSizeParameter(active_alfps, active_alfps->alf_block_size_idx);
  }

  // set temp_alfps
  SetALFParameters(temp_alfps, active_alfps);
}

/*!
 *****************************************************************************************
 * \brief
 *    decision of filter coefficients by re-design
 *****************************************************************************************
 */
void ReDesignFilterDecision(imgpel** ImgOrg, imgpel** ImgDec)
{
  int i, not_balf_flag;
  int tmp_tap, tmp_num_coef, *tmp_coef;

  get_mem1Dint(&(tmp_coef), MAX_NUM_COEF);

  // store default alf parameters
  tmp_tap      = temp_alfps->alf.tap;
  tmp_num_coef = temp_alfps->alf.num_coeff;
  memcpy(tmp_coef, temp_alfps->alf.coeff, sizeof(int)*tmp_num_coef);

  for(i=0; i<NUM_OF_REDESIGN; i++)
  {
    ReDesignFilterCoeff();

    FilteringProcess(ImgDec, imgY_temp, temp_alfps->alf.coeff, temp_alfps->alf.tap);

    // set block-based on/off control flag
    not_balf_flag = SetALFBlockFlag(temp_alfps, ImgOrg, enc_picture->imgY, imgY_temp);
    if(not_balf_flag) continue;

    CalcRDCostBlock(ImgOrg, imgY_temp, enc_picture->imgY, temp_alfps, &rate, &dist, &cost);

    if( cost < min_cost )
    {
      min_rate = rate;
      min_dist = dist;
      min_cost = cost;
      SetALFParameters(active_alfps, temp_alfps);
      SetRestFrame(imgY_temp, imgY_best, img->width, img->height);
    }
  }

  // return default alf parameters
  temp_alfps->alf.tap       = tmp_tap;
  temp_alfps->alf.num_coeff = tmp_num_coef;
  memcpy(temp_alfps->alf.coeff, tmp_coef, sizeof(int)*tmp_num_coef);

  free_mem1Dint(tmp_coef);
}

/*!
 *****************************************************************************************
 * \brief
 *    re-design filter coefficients
 *****************************************************************************************
 */
void ReDesignFilterCoeff()
{  
  int    i, j, k, tap, N, err_code;
  int    x, y, bn, bx, by, max_bx, max_by;
  double **corr, *h;
  int    *qh;
  AlfBlock *currAlfBlk;
  unsigned int  ****tmp_blk_corr;

  corr = correlation[0];
  h    = dbl_alf_coef;
  qh   = temp_alfps->alf.coeff;
  tap  = temp_alfps->alf.tap;
  N    = temp_alfps->alf.num_coeff;
  bn   = temp_alfps->alf_block_size / alf_blk_size[0];

  switch(tap)
  {
  case 5:
    tmp_blk_corr = blk_corr5x5;
    break;
  case 7:
    tmp_blk_corr = blk_corr7x7;
    break;
  case 9:
    tmp_blk_corr = blk_corr;
    break;
  default:
    printf("not support filter size : %d", tap);
    exit(1);
    break;
  }

  // initialize correration
  for(i=0; i<N; i++)
    memset(corr[i], 0, sizeof(double)*(N+1));

  for (i=0; i<temp_alfps->num_alf_block; i++)
  {
    currAlfBlk = &temp_alfps->alf_blk_data[i];
    if(currAlfBlk->alf_blk_flag)
    {
      by = (i / temp_alfps->WidthInAlfBlks) * bn;
      bx = (i % temp_alfps->WidthInAlfBlks) * bn;
      max_by = by + bn;
      max_bx = bx + bn;
      if(max_by>img->max_HeightInAlfBlks) max_by = img->max_HeightInAlfBlks;
      if(max_bx>img->max_WidthInAlfBlks ) max_bx = img->max_WidthInAlfBlks;
      for(y=by; y<max_by; y++)
      {
        for(x=bx; x<max_bx; x++)
        {
          for(j=0; j<N; j++)
          {
            for(k=j; k<N+1; k++)
            {
              corr[j][k] += (double)tmp_blk_corr[y][x][j][k-j];
            }
          }
        }
      }
    }
  }

  for(j=0; j<N-1; j++)
    for(k=j+1; k<N; k++)
      corr[k][j] = corr[j][k];

  err_code = Gauss(corr,N);
  if(err_code)
  {
    ClearFilterCoefInt(qh, N);
  }
  else
  {
    for(i=0; i<N; i++)
      h[i] = corr[i][N];
    QuantFilterCoef(h, qh, tap, img->bitdepth_luma);
  }
}

/*!
 *****************************************************************************************
 * \brief
 *    set parameters for Block-based ALF
 *****************************************************************************************
 */
void SetBlockSizeParameter(alf_parameter_set_rbsp_t *alfps, int size_idx)
{
  alfps->alf_block_size_idx = size_idx;
  alfps->alf_block_size     = alf_blk_size[size_idx];
  alfps->WidthInAlfBlks     = img->width  / alfps->alf_block_size;
  alfps->HeightInAlfBlks    = img->height / alfps->alf_block_size;
  if( (img->width  % alfps->alf_block_size)!=0 ) alfps->WidthInAlfBlks++;
  if( (img->height % alfps->alf_block_size)!=0 ) alfps->HeightInAlfBlks++;
  alfps->num_alf_block      = alfps->WidthInAlfBlks * alfps->HeightInAlfBlks;

  // set image parameter for CABAC
  img->alf_block_size  = alfps->alf_block_size;
  img->WidthInAlfBlks  = alfps->WidthInAlfBlks;
  img->HeightInAlfBlks = alfps->HeightInAlfBlks;
  img->num_alf_block   = alfps->num_alf_block;
  img->alf_blk_data    = alfps->alf_blk_data;
}

/*!
 *****************************************************************************************
 * \brief
 *    set on/off flag per block
 *****************************************************************************************
 */
int SetALFBlockFlag(alf_parameter_set_rbsp_t *alfps, imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest)
{
  int i;
  int64 diff_dec, diff_rest;
  AlfBlock *currAlfBlk;
  int all_blk_flag_zero=1;

  for (i=0; i<alfps->num_alf_block; i++)
  { 
    img->current_alfb_nr = i;
    CheckAvailabilityOfNeighborsAlf();

    currAlfBlk = &alfps->alf_blk_data[i];

    diff_dec  = FindDistortionBlock(i, ImgOrg, ImgDec,  alfps->alf_block_size, alfps->WidthInAlfBlks);
    diff_rest = FindDistortionBlock(i, ImgOrg, ImgRest, alfps->alf_block_size, alfps->WidthInAlfBlks);

    if(diff_rest < diff_dec)
    {
      all_blk_flag_zero = 0;
      currAlfBlk->alf_blk_flag = TRUE;
    }
    else
    {
      currAlfBlk->alf_blk_flag = FALSE;
    }
  }

  return all_blk_flag_zero;
}

/*!
 *****************************************************************************************
 * \brief
 *    copy from dec block to rest block
 *****************************************************************************************
 */
void CopyDecToRestBlock(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int blk_size, int width_in_alf_blks)
{
  int x, y, blk_x, blk_y, max_blk_x, max_blk_y;

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, width_in_alf_blks);

  max_blk_x = blk_x + blk_size;
  max_blk_y = blk_y + blk_size;
  if(max_blk_x > img->width)  max_blk_x = img->width;
  if(max_blk_y > img->height) max_blk_y = img->height;

  for(y=blk_y; y<max_blk_y; y++)
  {
    for(x=blk_x; x<max_blk_x; x++)
    {
      ImgRest[y][x] = ImgDec[y][x];
    }
  }
}

/*!
 ************************************************************************
 * \brief
 *    returns the x and y block coordinates for a given AlfBlockAddress
 ************************************************************************
 */
void GetBlockPosForBALF(int blk_addr, int *x, int*y, int width_in_blks)
{
  *x = (blk_addr % width_in_blks);
  *y = (blk_addr / width_in_blks);
}

/*!
 ************************************************************************
 * \brief
 *    returns the x and y sample coordinates for a given AlfBlockAddress
 ************************************************************************
 */
void GetSamplePosForBALF(int blk_addr, int *x, int*y, int blk_size, int width_in_blks)
{
  GetBlockPosForBALF(blk_addr, x, y, width_in_blks);
  
  (*x) *= blk_size;
  (*y) *= blk_size;
}

/*!
 ************************************************************************
 * \brief
 *    Allocate memory array for blk_corr -> unsigned int array4D[height][width][rows][columns]
 *
 * \par Output:
 *    memory size in bytes
 ************************************************************************
 */
int get_mem_blk_corr(unsigned int *****blk_corr, int height, int width, int rows, int columns )
{
  int  i, j, k;
  int  num;       // num of coeff (column + column-1 +  column-2 + ... + 2)
  unsigned int ****ptr_4D, ***ptr_3D;

  num = rows*(columns+2)/2;

  if((*blk_corr = (unsigned int ****) calloc(sizeof(unsigned int***), height))==NULL)
    no_mem_exit("get_mem_blk_corr: blk_corr");

  for(i=0; i<height; i++)
  {
    ptr_4D = *blk_corr+i;
    if((*ptr_4D = (unsigned int ***) calloc(sizeof(unsigned int**), width))==NULL)
      no_mem_exit("get_mem_blk_corr: blk_corr");
    for(j=0; j<width; j++)
    {
      ptr_3D = *ptr_4D+j;
      if((*ptr_3D = (unsigned int **) calloc(sizeof(unsigned int*), rows))==NULL)
        no_mem_exit("get_mem_blk_corr: blk_corr");
      if(((*ptr_3D)[0] = (unsigned int *) calloc(sizeof(unsigned int), num))==NULL)
        no_mem_exit("get_mem_blk_corr: blk_corr");
      for(k=0; k<rows-1; k++)
        (*ptr_3D)[k+1] = (*ptr_3D)[k] + columns - k;
    }
  }

  return height*width*num*sizeof(unsigned int);
}

/*!
 ************************************************************************
 * \brief
 *    free 4D memory array
 *    which was alocated with get_mem_blk_corr()
 ************************************************************************
 */
void free_mem_blk_corr(unsigned int ****blk_corr, int height, int width )
{
  int  i, j;

  if (blk_corr)
  {
    for(j=0;j<height;j++)
    {
      if (blk_corr[j])
      {
        for (i=0;i<width;i++)
        { 
          if (blk_corr[j][i])
          {
            if (blk_corr[j][i][0]) 
              free (blk_corr[j][i][0]);
            else
              error ("free_mem_blk_corr: trying to free unused memory",100);

            free (blk_corr[j][i]);
          }
          else
          {
            error ("free_mem_blk_corr: trying to free unused memory",100);
          }
        }
        free (blk_corr[j]);
      }
      else
      {
        error ("free_mem_blk_corr: trying to free unused memory",100);
      }
    }
    free (blk_corr);
  } else
  {
    error ("free_mem_blk_corr: trying to free unused memory",100);
  }
}

/*!
 ************************************************************************
 * \brief
 *    returns 1 if the ALF block at the given address is available
 ************************************************************************
 */
int ALFBlockIsAvailable(int blk_addr)
{
  if ((blk_addr < 0) || (blk_addr > ((int)img->num_alf_block - 1)))
    return 0;
  
  return 1;
}

/*!
 ************************************************************************
 * \brief
 *    Checks the availability of neighboring ALF blocks of
 *    the current ALF block for prediction and context determination;
 ************************************************************************
 */
void CheckAvailabilityOfNeighborsAlf()
{
  const int alfb_nr  = img->current_alfb_nr;
  AlfBlock *currAlfB = &img->alf_blk_data[alfb_nr];

  // mark all neighbors as unavailable
  currAlfB->alfb_available_up   = NULL;
  currAlfB->alfb_available_left = NULL;

  currAlfB->alfbAddrA = alfb_nr - 1;
  currAlfB->alfbAddrB = alfb_nr - img->WidthInAlfBlks;
  currAlfB->alfbAddrC = alfb_nr - img->WidthInAlfBlks + 1;
  currAlfB->alfbAddrD = alfb_nr - img->WidthInAlfBlks - 1;

  currAlfB->alfbAvailA = ALFBlockIsAvailable(currAlfB->alfbAddrA) && (( alfb_nr    % img->WidthInAlfBlks)!=0);
  currAlfB->alfbAvailB = ALFBlockIsAvailable(currAlfB->alfbAddrB);
  currAlfB->alfbAvailC = ALFBlockIsAvailable(currAlfB->alfbAddrC) && (((alfb_nr+1) % img->WidthInAlfBlks)!=0);
  currAlfB->alfbAvailD = ALFBlockIsAvailable(currAlfB->alfbAddrD) && (( alfb_nr    % img->WidthInAlfBlks)!=0);

  if (currAlfB->alfbAvailA) currAlfB->alfb_available_left = &(img->alf_blk_data[currAlfB->alfbAddrA]);
  if (currAlfB->alfbAvailB) currAlfB->alfb_available_up   = &(img->alf_blk_data[currAlfB->alfbAddrB]);
}

/*!
 ************************************************************************
 * \brief
 *    get neighbouring positions for non-aff coding
 * \param curr_alfb_nr
 *   current ALF block number (decoding order)
 * \param xN
 *    input x position
 * \param yN
 *    input y position
 * \param pix
 *    returns position informations
 ************************************************************************
 */
void GetNonAffNeighbourAlfBlock(unsigned int curr_alfb_nr, int xN, int yN, PixelPos *pix)
{
  AlfBlock *currAlfB = &img->alf_blk_data[curr_alfb_nr];
  int maxW, maxH;

  maxW = img->alf_block_size;
  maxH = img->alf_block_size;

  if ((xN<0)&&(yN<0))
  {
    pix->mb_addr   = currAlfB->alfbAddrD;
    pix->available = currAlfB->alfbAvailD;
  }
  else if ((xN<0)&&((yN>=0)&&(yN<maxH)))
  {
    pix->mb_addr   = currAlfB->alfbAddrA;
    pix->available = currAlfB->alfbAvailA;
  }
  else if (((xN>=0)&&(xN<maxW))&&(yN<0))
  {
    pix->mb_addr   = currAlfB->alfbAddrB;
    pix->available = currAlfB->alfbAvailB;
  }
  else if (((xN>=0)&&(xN<maxW))&&((yN>=0)&&(yN<maxH)))
  {
    pix->mb_addr   = curr_alfb_nr;
    pix->available = 1;
  }
  else if ((xN>=maxW)&&(yN<0))
  {
    pix->mb_addr   = currAlfB->alfbAddrC;
    pix->available = currAlfB->alfbAvailC;
  }
  else 
  {
    pix->available = 0;
  }

  if (pix->available)
  {
    //pix->x = (xN + maxW) % maxW;
    //pix->y = (yN + maxH) % maxH;
    pix->x = (xN + maxW) & (maxW - 1);
    pix->y = (yN + maxH) & (maxH - 1);

    GetSamplePosForBALF(pix->mb_addr, &(pix->pos_x), &(pix->pos_y), img->alf_block_size, img->WidthInAlfBlks);

    pix->pos_x += pix->x;
    pix->pos_y += pix->y;
  }
}

/*!
 ************************************************************************
 * \brief
 *    get neighbouring positions. MB AFF is automatically used from img structure
 * \param curr_alfb_nr
 *   current ALF block number (decoding order)
 * \param xN
 *    input x position
 * \param yN
 *    input y position
 * \param pix
 *    returns position informations
 ************************************************************************
 */
void GetNeighbourAlfBlock(int curr_alfb_nr, int xN, int yN, PixelPos *pix)
{
  if (curr_alfb_nr<0)
    error ("GetNeighbourAlfBlock: invalid block number", 100);

  GetNonAffNeighbourAlfBlock(curr_alfb_nr, xN, yN, pix);
}


//------- send ALF parameters
/*!
 *****************************************************************************************
 * \brief
 *    send ALF parameters
 *****************************************************************************************
 */
int send_ALFParameterSet(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream)
{
  int new_bitstream_flag;
  int len = 0;

  new_bitstream_flag = 0;
  if (bitstream==NULL)
  {
    // allocate bitstream
    bitstream =  (Bitstream *) calloc(1, sizeof(Bitstream));
    bitstream->streamBuffer = (byte *) calloc(MAX_ALF_PARAM_SIZE, sizeof(byte));
    bitstream->bits_to_go=8;
    bitstream->byte_buf=0;
    bitstream->byte_pos=0;
    new_bitstream_flag = 1;
  }

  // ALF parameters

  // filter parameters for luma
  len += send_ALFLumaParam(alfps, bitstream);

  // filter parameters for chroma
  len += send_ALFChromaParam(alfps, bitstream);

  // block control parameters for luma
  len += send_ALFBlockControlParam(alfps, bitstream);

  // padding
  if (bitstream->bits_to_go != 8 )
  {
    while(bitstream->bits_to_go != 8)
    {
      len += u_1("ALFPS: padding bits", 1, bitstream); 
    }
  }

  if (new_bitstream_flag)
  {
    // free bitstream
    free(bitstream->streamBuffer);
    bitstream->streamBuffer=NULL;
    free(bitstream);
    bitstream=NULL;
  }

  return len;
}

/*!
 *****************************************************************************************
 * \brief
 *    send ALF parameters for luma
 *****************************************************************************************
 */
int send_ALFLumaParam(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream)
{
  int len = 0;
  int pos;
  char string[100];

  len += ue_v("ALFPS: alf_tap_size", (alfps->alf.tap-5)/2, bitstream);

  // filter coefficients for luma
  for(pos=0; pos<alfps->alf.num_coeff; pos++)
  {
    sprintf(string, "ALFPS: %d pos", pos);
    len += se_v(string, alfps->alf.coeff[pos], bitstream);
  }

  return len;
}

/*!
 *****************************************************************************************
 * \brief
 *    send ALF parameters for chroma
 *****************************************************************************************
 */
int send_ALFChromaParam(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream)
{
  int len = 0;
  int pos;
  char string[100];

  len += u_v(2, "ALFPS: alf_chroma_idc", alfps->chroma_idc, bitstream);

  if(alfps->chroma_idc)
  {
    len += ue_v("ALFPS: filter_tap_chroma", (alfps->alf.tap_chroma-5)/2, bitstream);

    // filter coefficients for chroma
    for(pos=0; pos<alfps->alf.num_coeff_chroma; pos++)
    {
      sprintf(string, "ALFPS: [chroma] %d pos", pos);
      len += se_v(string, alfps->alf.coeff_chroma[pos], bitstream);
    }
  }

  return len;
}

/*!
 *****************************************************************************************
 * \brief
 *    send ALF parameters for block-based
 *****************************************************************************************
 */
int send_ALFBlockControlParam(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream)
{
  int len = 0;

  len += u_1("ALFPS: block_control_flag", alfps->block_control_flag, bitstream);

  if(alfps->block_control_flag)
  {
    len += u_v(3, "ALFPS: alf_block_size_idx", alfps->alf_block_size_idx, bitstream);

    // block-based on/off flag
    if (input->symbol_mode == CABAC)
    {
      len += send_BlockFlagCABAC(alfps, bitstream);
    }
    else
    {
      len += send_BlockFlagRunLength(alfps, bitstream);
    }
  }

  return len;
}

/*!
 *****************************************************************************************
 * \brief
 *    send block flag by Run-length coding
 *****************************************************************************************
 */
int send_BlockFlagRunLength(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream)
{
  int i;
  int len = 0;
  int alf_blk_flag_run=0; 
  int prev_alf_blk_flag=0;
  AlfBlock *currAlfBlk;

  alf_blk_flag_run = 0;

  for (i=0; i<alfps->num_alf_block; i++)
  {
    currAlfBlk = &alfps->alf_blk_data[i];

    if( i==0 )
    {
      len += u_1("ALFPS: alf_blk_flag", currAlfBlk->alf_blk_flag, bitstream);
      alf_blk_flag_run = 0;
      prev_alf_blk_flag = currAlfBlk->alf_blk_flag;
    }
    else
    {
      if( (prev_alf_blk_flag==currAlfBlk->alf_blk_flag) && (i!=alfps->num_alf_block-1) )
      {
        alf_blk_flag_run++;
      }
      else
      {
        if( (prev_alf_blk_flag==currAlfBlk->alf_blk_flag) && (i==alfps->num_alf_block-1) )
        {
          alf_blk_flag_run++;
        }

        len += ue_v("ALFPS: alf_blk_flag_run", alf_blk_flag_run, bitstream);
        alf_blk_flag_run = 0;
        prev_alf_blk_flag = currAlfBlk->alf_blk_flag;
      }
    }
  }

  return len;
}

/*!
 *****************************************************************************************
 * \brief
 *    send block flag by CABAC
 *****************************************************************************************
 */
int send_BlockFlagCABAC(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream)
{
  int i;
  int len = 0;
  AlfBlock *currAlfBlk;
  SyntaxElement *currSE;
  DataPartition *dataPart;
  EncodingEnvironmentPtr eep=0;
  extern int cabac_encoding;
  Slice *currSlice = img->currentSlice;
#if TRACE
  char string[100];
#endif

  if( ( dataPart = (DataPartition *) calloc(1, sizeof(DataPartition)) ) == NULL )
    no_mem_exit ("calloc DataPartition");
  dataPart->bitstream = bitstream;
  dataPart->ee_cabac  = (currSlice->partArr[0]).ee_cabac;
  eep = &(dataPart->ee_cabac);
  writeVlcByteAlign(bitstream);
  arienco_start_encoding_AlfBlockFlag(eep, bitstream->streamBuffer, &(bitstream->byte_pos));
  init_contexts();
  dataPart->writeSyntaxElement = writeSyntaxElement_CABAC;
  cabac_encoding = 1;

  for (i=0; i<alfps->num_alf_block; i++)
  {
    currAlfBlk = &alfps->alf_blk_data[i];

    img->current_alfb_nr = i;
    currAlfBlk->currSEnr = 0;
    currSE = &img->MB_SyntaxElements[currAlfBlk->currSEnr];

    CheckAvailabilityOfNeighborsAlfCABAC(alfps);
    currSE->value1  = currAlfBlk->alf_blk_flag;
    currSE->type    = SE_HEADER;
    currSE->writing = writeAlfBlockFlagCABAC;
    dataPart->writeSyntaxElement(currSE, dataPart);
#if TRACE
    currSE->bitpattern = eep->Ebuffer;
    sprintf(string, "ALFPS: alf_blk_flag[%4d]", i);
    snprintf(currSE->tracestring, TRACESTRING_SIZE, string);
    trace2out(currSE);
#endif
    len += currSE->len;
    currAlfBlk->currSEnr++;

    // terminate the arithmetic code
    if (i == alfps->num_alf_block-1)
    {
      biari_encode_symbol_final(eep, 1); 
#if TRACE
      fprintf (p_trace, "      CABAC terminating bit = %d\n",1);
#endif
      arienco_done_encoding(eep);
    }
  }
  free(dataPart);
  cabac_encoding = 0;

  return len;
}

/*!
 *****************************************************************************************
 * \brief
 *    terminate slice for ALF
 *****************************************************************************************
 */
int TerminateSliceForALF(Picture* pic)
{
  static int MbWidthC  [4]= { 0, 8, 8,  16};
  static int MbHeightC [4]= { 0, 8, 16, 16};
  
  int bytes_written;
  Bitstream *currStream;
  Slice *currSlice;
  EncodingEnvironmentPtr eep;
  int i, j, slice;
  int byte_pos_before_startcode_emu_prevention;
  int min_num_bytes=0;
  int stuffing_bytes=0;
  int RawMbBits;

  int sh_end_pos, alf_end_pos, sd_end_pos;

  for (slice=0; slice<pic->no_slices; slice++)
  {
    currSlice   = pic->slices[slice];
    sh_end_pos  = currSlice->sh_byte_pos;
    alf_end_pos = sh_end_pos + alfstream->byte_pos;
    for (i=0; i<currSlice->max_part_nr; i++)
    {
      currStream = (currSlice->partArr[i]).bitstream;
      if(active_alfps->alf_flag)
      {
        sd_end_pos = currStream->byte_pos;
        for (j=0; j<sd_end_pos - sh_end_pos; j++)
        {
          currStream->streamBuffer[sd_end_pos+alfstream->byte_pos-j-1] = currStream->streamBuffer[sd_end_pos-j-1];
        }
        for (j=0; j<alfstream->byte_pos; j++)
        {
          currStream->streamBuffer[sh_end_pos+j] = alfstream->streamBuffer[j];
        }
        currStream->byte_pos = sd_end_pos + alfstream->byte_pos;
      }

      if (input->symbol_mode == UVLC)
      {
        SODBtoRBSP(currStream);
        byte_pos_before_startcode_emu_prevention = currStream->byte_pos;
        currStream->byte_pos = RBSPtoEBSP(currStream->streamBuffer, 0 , currStream->byte_pos, 0);
        *(stats->em_prev_bits) += (currStream->byte_pos - byte_pos_before_startcode_emu_prevention) * 8;
      }
      else     // CABAC
      {
        eep = &((currSlice->partArr[i]).ee_cabac);
        // terminate the arithmetic code
        arienco_done_encoding(eep);
        currStream->bits_to_go = eep->Ebits_to_go;
        currStream->byte_buf = 0;
        bytes_written = currStream->byte_pos;
        img->bytes_in_picture += currStream->byte_pos;

        byte_pos_before_startcode_emu_prevention= currStream->byte_pos;
        if (i==((currSlice->max_part_nr-1)))
        {
          RawMbBits = 256 * img->bitdepth_luma + 2 * MbWidthC[active_sps->chroma_format_idc] * MbHeightC[active_sps->chroma_format_idc] * img->bitdepth_chroma;
          min_num_bytes = ((96 * get_pic_bin_count()) - (RawMbBits * (int)img->PicSizeInMbs *3) + 1023) / 1024;
          if (min_num_bytes>img->bytes_in_picture)
          {
            stuffing_bytes = min_num_bytes - img->bytes_in_picture;
            printf ("CABAC stuffing words = %6d\n", stuffing_bytes/3);
          }
        }

        currStream->byte_pos = RBSPtoEBSP(currStream->streamBuffer, 0, currStream->byte_pos, currStream->byte_pos + stuffing_bytes);
        *(stats->em_prev_bits) += (currStream->byte_pos - byte_pos_before_startcode_emu_prevention) * 8;
      }           // CABAC
    }           // partition loop
  }

  alfstream->byte_pos  = 0;

  return 0;
}


#endif // ADAPTIVE_LOOP_FILTER

